@@ -1,7 +1,8 @@
 /*
- * $Id: shoot.c,v 1.9 2003/12/30 21:44:27 calrissian Exp $
+ * $Id: shoot.c,v 1.32 2004/07/15 13:21:12 calrissian Exp $
  *
- * Copyright (C) 2002-2003 Fhg Fokus
+ * Copyright (C) 2002-2004 Fhg Fokus
+ * Copyright (C) 2004 Nils Ohlmeier
  *
  * This file belongs to sipsak, a free sip testing tool.
  *
@@ -20,19 +21,32 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <limits.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <regex.h>
 #include <sys/socket.h>
 #include <sys/poll.h>
 #include <netinet/in.h>
+#include <arpa/inet.h>
 
 #include "shoot.h"
+
+#ifdef RAW_SUPPORT
+#include <netinet/in_systm.h>
+#include <netinet/ip.h>
+#include <netinet/ip_icmp.h>
+
+#define __FAVOR_BSD
+#include <netinet/udp.h>
+#endif
+
 #include "sipsak.h"
 #include "request.h"
 #include "auth.h"
 #include "header_f.h"
 #include "helper.h"
+#include "exit_code.h"
 
 #ifndef DEFAULT_RETRYS
 #define DEFAULT_RETRYS 5
@@ -53,51 +67,103 @@ bouquets and brickbats to farhan@hotfoon.com
 void shoot(char *buff)
 {
 	struct sockaddr_in	addr;
-	struct timeval	tv, sendtime, recvtime, firstsendt, delaytime;
+	struct timeval	tv, sendtime, recvtime, firstsendt, delaytime, starttime;
 	struct timezone tz;
 	struct timespec sleep_ms_s, sleep_rem;
-	struct pollfd sockerr;
+	struct pollfd 	sockerr;
 	int redirected, retryAfter, nretries;
-	int sock , i, len, ret, usrlocstep;
+	int usock, csock, i, len, ret;
 	int dontsend, cseqtmp, rand_tmp;
 	int rem_rand, retrans_r_c, retrans_s_c;
 	int randretrys = 0;
 	int cseqcmp = 0;
 	int rem_namebeg = 0;
-	double big_delay, tmp_delay;
+	double big_delay, tmp_delay, senddiff;
 	char *contact, *foo, *bar, *lport_str;
 	char *crlf = NULL;
 	char reply[BUFSIZE];
 	fd_set	fd;
 	socklen_t slen;
 	regex_t redexp, proexp, okexp, tmhexp, errexp, authexp;
+	enum usteps { REG_REP, INV_RECV, INV_OK_RECV, INV_ACK_RECV, MES_RECV, 
+					MES_OK_RECV, UNREG_REP};
+	enum usteps usrlocstep = REG_REP;
+#ifdef RAW_SUPPORT
+	struct sockaddr_in faddr;
+	struct ip 		*r_ip_hdr, *s_ip_hdr;
+	struct icmp 	*icmp_hdr;
+	struct udphdr 	*udp_hdr;
+	size_t r_ip_len, s_ip_len, icmp_len;
+	int srcport, dstport, rawsock;
+	unsigned int flen;
+	char fstr[INET_ADDRSTRLEN];
+#endif
 
 	/* the vars are filled by configure */
 	nretries = DEFAULT_RETRYS;
-	retryAfter = DEFAULT_TIMEOUT;
+	/* retryAfter = DEFAULT_TIMEOUT; */
+	retryAfter = SIP_T1;
 
 	/* initalize some local vars */
 	redirected = 1;
-	usrlocstep=dontsend=retrans_r_c=retrans_s_c = 0;
+	dontsend=retrans_r_c=retrans_s_c = 0;
 	big_delay=tmp_delay = 0;
 	delaytime.tv_sec = 0;
 	delaytime.tv_usec = 0;
 
-	/* create the socket */
-	sock = (int)socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
-	if (sock==-1) {
-		perror("socket creation failed");
-		exit(2);
+	/* create the un-connected socket */
+	usock = (int)socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
+	if (usock==-1) {
+		perror("unconnected UDP socket creation failed");
+		exit_code(2);
 	}
 
+	memset(&addr, 0, sizeof(addr));
 	addr.sin_family=AF_INET;
 	addr.sin_addr.s_addr = htonl( INADDR_ANY );
 	addr.sin_port = htons((short)lport);
-	if (bind( sock, (struct sockaddr *) &addr, sizeof(addr) )==-1) {
-		perror("socket binding failed");
-		exit(2);
+	if (bind( usock, (struct sockaddr *) &addr, sizeof(addr) )==-1) {
+		perror("unconnected UDP socket binding failed");
+		exit_code(2);
 	}
 
+	/* for the via line we need our listening port number */
+	if (lport==0){
+		memset(&addr, 0, sizeof(addr));
+		slen=sizeof(addr);
+		getsockname(usock, (struct sockaddr *)&addr, &slen);
+		lport=ntohs(addr.sin_port);
+	}
+
+#ifdef RAW_SUPPORT
+	/* try to create the raw socket */
+	rawsock = (int)socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
+	if (rawsock==-1) {
+		if (verbose)
+			printf("Warning: need raw socket (root privileges) to receive all ICMP errors\n");
+#endif
+		/* create the connected socket as a primitve alternative to the 
+		   raw socket*/
+		csock = (int)socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
+		if (csock==-1) {
+			perror("connected UDP socket creation failed");
+			exit_code(2);
+		}
+
+		addr.sin_family=AF_INET;
+		addr.sin_addr.s_addr = htonl( INADDR_ANY );
+		addr.sin_port = htons((short)0);
+		if (bind( csock, (struct sockaddr *) &addr, sizeof(addr) )==-1) {
+			perror("connected UDP socket binding failed");
+			exit_code(2);
+		}
+#ifdef RAW_SUPPORT
+	}
+	else {
+		csock = -1;
+	}
+#endif
+
 	if (sleep_ms != 0) {
 		if (sleep_ms == -2) {
 			rand_tmp = rand();
@@ -110,14 +176,6 @@ void shoot(char *buff)
 		}
 	}
 
-	/* for the via line we need our listening port number */
-	if ((via_ins||usrloc||invite||message||replace_b) && lport==0){
-		memset(&addr, 0, sizeof(addr));
-		slen=sizeof(addr);
-		getsockname(sock, (struct sockaddr *)&addr, &slen);
-		lport=ntohs(addr.sin_port);
-	}
-
 	if (replace_b){
 		replace_string(buff, "$dsthost$", domainname);
 		replace_string(buff, "$srchost$", fqdn);
@@ -153,19 +211,22 @@ void shoot(char *buff)
 			else if (message)
 				nretries=3*(nameend-namebeg)+3;
 			else
-				nretries=nameend-namebeg+1;
+				nretries=2*(nameend-namebeg)+2;
 			create_msg(buff, REQ_REG);
-			usrlocstep=0;
+			usrlocstep=REG_REP;
 		}
 		else if (invite) {
 			nretries=3*(nameend-namebeg)+3;
 			create_msg(buff, REQ_INV);
-			usrlocstep=1;
+			usrlocstep=INV_RECV;
 		}
 		else {
 			nretries=2*(nameend-namebeg)+2;
 			create_msg(buff, REQ_MES);
-			usrlocstep=4;
+			if (mes_body)
+				usrlocstep=MES_OK_RECV;
+			else
+				usrlocstep=MES_RECV;
 		}
 		cseqcmp=1;
 	}
@@ -182,7 +243,7 @@ void shoot(char *buff)
 	}
 	else if (flood){
 		/* this should be the max of an (32 bit) int without the sign */
-		if (namebeg==-1) namebeg=2147483647;
+		if (namebeg==-1) namebeg=INT_MAX;
 		nretries=namebeg;
 		namebeg=1;
 		create_msg(buff, REQ_FLOOD);
@@ -208,7 +269,7 @@ void shoot(char *buff)
 			namebeg=1;
 			create_msg(buff, REQ_OPT);
 		}
-		retryAfter = retryAfter / 10;
+		/* retryAfter = retryAfter / 10; */
 		if(maxforw!=-1)
 			set_maxforw(buff);
 		if(via_ins)
@@ -223,16 +284,17 @@ void shoot(char *buff)
 
 		/* destination socket init here because it could be changed in a 
 		   case of a redirect */
+		memset(&addr, 0, sizeof(addr));
 		addr.sin_addr.s_addr = address;
 		addr.sin_port = htons((short)rport);
 		addr.sin_family = AF_INET;
 	
-		/* we connect as per the RFC 2543 recommendations
-		   modified from sendto/recvfrom */
-		ret = connect(sock, (struct sockaddr *)&addr, sizeof(addr));
-		if (ret==-1) {
-			perror("no connect");
-			exit(2);
+		if (csock != -1) {
+			ret = connect(csock, (struct sockaddr *)&addr, sizeof(addr));
+			if (ret==-1) {
+				perror("connecting UDP socket failed");
+				exit_code(2);
+			}
 		}
 
 		/* here we go for the number of nretries which strongly depends on the 
@@ -245,36 +307,39 @@ void shoot(char *buff)
 			/* some initial output */
 			else if ((usrloc||invite||message) && (verbose > 1) && !dontsend) {
 				switch (usrlocstep) {
-					case 0:
+					case REG_REP:
 						if (nameend>0)
 							printf("registering user %s%i... ", username, 
 								namebeg);
 						else
 							printf("registering user %s... ", username);
 						break;
-					case 1:
+					case INV_RECV:
 						if (nameend>0)
 							printf("inviting user %s%i... ", username, namebeg);
 						else
 							printf("invitng user %s... ", username);
 						break;
-					case 2:
+					case INV_OK_RECV:
 						printf("sending invite reply... ");
 						break;
-					case 3:
+					case INV_ACK_RECV:
 						printf("sending invite ack... ");
 						break;
-					case 4:
+					case MES_RECV:
 						if (nameend>0)
 							printf("sending message to %s%i... ", username,
 								namebeg);
 						else
 							printf("sending message to %s... ", username);
 						break;
-					case 5:
-						printf("sending message reply... ");
+					case MES_OK_RECV:
+						if (mes_body)
+							printf("sending message ... \n");
+						else
+							printf("sending message reply... ");
 						break;
-					case 6:
+					case UNREG_REP:
 						if (nameend>0)
 							printf("remove binding for %s%i...", username, 
 								namebeg);
@@ -304,12 +369,17 @@ void shoot(char *buff)
 
 			if (! dontsend) {
 				/* lets fire the request to the server and store when we did */
-				ret = send(sock, buff, strlen(buff), 0);
+				if (csock == -1) {
+					ret = sendto(usock, buff, strlen(buff), 0, (struct sockaddr *)&addr, sizeof(addr));
+				}
+				else {
+					ret = send(csock, buff, strlen(buff), 0);
+				}
 				(void)gettimeofday(&sendtime, &tz);
 				if (ret==-1) {
 					printf("\n");
 					perror("send failure");
-					exit(2);
+					exit_code(2);
 				}
 			}
 			else {
@@ -324,7 +394,13 @@ void shoot(char *buff)
 				tv.tv_usec = (retryAfter % 1000) * 1000;
 
 				FD_ZERO(&fd);
-				FD_SET(sock, &fd); 
+				FD_SET(usock, &fd); 
+				if (csock != -1)
+					FD_SET(csock, &fd); 
+#ifdef RAW_SUPPORT
+				if (rawsock != -1)
+					FD_SET(rawsock, &fd); 
+#endif
 
 				ret = select(FD_SETSIZE, &fd, NULL, NULL, &tv);
 				(void)gettimeofday(&recvtime, &tz);
@@ -333,18 +409,26 @@ void shoot(char *buff)
 					/* store the time of our first send */
 					if (i==0)
 						memcpy(&firstsendt, &sendtime, sizeof(struct timeval));
+					if (retryAfter == SIP_T1)
+						memcpy(&starttime, &sendtime, sizeof(struct timeval));
 					/* lets see if we at least received an icmp error */
-					sockerr.fd=sock;
+					if (csock == -1) 
+						sockerr.fd=usock;
+					else
+						sockerr.fd=csock;
 					sockerr.events=POLLERR;
 					if ((poll(&sockerr, 1, 10))==1) {
 						if (sockerr.revents && POLLERR) {
-							recv(sock, reply, strlen(reply), 0);
+							if (csock == -1)
+								recv(usock, reply, strlen(reply), 0);
+							else
+								recv(csock, reply, strlen(reply), 0);
 							printf("\n");
 							perror("send failure");
 							if (randtrash) 
 								printf ("last message before send failure:"
 									"\n%s\n", buff);
-							exit(3);
+							exit_code(3);
 						}
 					}
 					/* printout that we did not received anything */
@@ -368,7 +452,7 @@ void shoot(char *buff)
 									"times without getting a response:\n%s\n"
 									"give up further retransmissions...\n", 
 									buff);
-								exit(3);
+								exit_code(3);
 							}
 							else {
 								printf("resending it without additional "
@@ -377,9 +461,18 @@ void shoot(char *buff)
 							}
 						}
 					}
-					retryAfter = retryAfter * 2;
-					if (retryAfter > DEFAULT_TIMEOUT) 
-						retryAfter = DEFAULT_TIMEOUT;
+					senddiff = deltaT(&starttime, &recvtime);
+					if (senddiff > 64 * SIP_T1) {
+						if (verbose)
+							printf("*** giving up, no response after %.3f ms\n",
+								senddiff);
+						exit_code(3);
+					}
+					/* set retry time according to RFC3261 */
+					if (retryAfter *2 < SIP_T2)
+						retryAfter = retryAfter * 2;
+					else
+						retryAfter = SIP_T2;
 					retrans_s_c++;
 					if (delaytime.tv_sec == 0)
 						memcpy(&delaytime, &sendtime, sizeof(struct timeval));
@@ -388,14 +481,20 @@ void shoot(char *buff)
 				}
 				else if ( ret == -1 ) {
 					perror("select error");
-					exit(2);
+					exit_code(2);
 				}
-				else if (FD_ISSET(sock, &fd)) {
+				else if (FD_ISSET(usock, &fd) || ((csock != -1) && FD_ISSET(csock, &fd))) {
 					/* no timeout, no error ... something has happened :-) */
 				 	if (!trace && !usrloc && !invite && !message && !randtrash 
 						&& (verbose > 1))
 						printf ("\nmessage received:\n");
 				}
+#ifdef RAW_SUPPORT
+				else if ((rawsock != -1) && FD_ISSET(rawsock, &fd)) {
+					if (verbose > 1)
+						printf("\nreceived ICMP packet");
+				}
+#endif
 				else {
 					printf("\nselect returned succesfuly, nothing received\n");
 					continue;
@@ -404,13 +503,70 @@ void shoot(char *buff)
 				/* we are retrieving only the extend of a decent 
 				   MSS = 1500 bytes */
 				len = sizeof(addr);
-				ret = recv(sock, reply, BUFSIZE, 0);
+				if (FD_ISSET(usock, &fd)) {
+					ret = recv(usock, reply, BUFSIZE, 0);
+				}
+				else if ((csock != -1) && (FD_ISSET(csock, &fd))) {
+					ret = recv(csock, reply, BUFSIZE, 0);
+				}
+#ifdef RAW_SUPPORT
+				else if ((rawsock != -1) && (FD_ISSET(rawsock, &fd))) {
+					/* lets check if the ICMP message matches with our 
+					   sent packet */
+					flen = sizeof(faddr);
+					memset(&faddr, 0, sizeof(addr));
+					ret = recvfrom(rawsock, reply, BUFSIZE, 0, (struct sockaddr *)&faddr, &flen);
+					if (ret == -1) {
+						perror("error while trying to read from icmp raw socket");
+						exit_code(2);
+					}
+					r_ip_hdr = (struct ip *) reply;
+					r_ip_len = r_ip_hdr->ip_hl << 2;
+
+					icmp_hdr = (struct icmp *) (reply + r_ip_len);
+					icmp_len = ret - r_ip_len;
+
+					if (icmp_len < 8) {
+						if (verbose > 1)
+							printf(": ignoring (ICMP header length below 8 bytes)\n");
+						continue;
+					}
+					else if (icmp_len < 36) {
+						if (verbose > 1)
+							printf(": ignoring (ICMP message too short to contain IP and UDP header)\n");
+						continue;
+					}
+					s_ip_hdr = (struct ip *) ((char *)icmp_hdr + 8);
+					s_ip_len = s_ip_hdr->ip_hl << 2;
+					if (s_ip_hdr->ip_p == IPPROTO_UDP) {
+						udp_hdr = (struct udphdr *) ((char *)s_ip_hdr + s_ip_len);
+						srcport = ntohs(udp_hdr->uh_sport);
+						dstport = ntohs(udp_hdr->uh_dport);
+						if ((srcport == lport) && (dstport == rport)) {
+							inet_ntop(AF_INET, &faddr.sin_addr, fstr, INET_ADDRSTRLEN);
+							printf(" (type: %u, code: %u): from %s\n", icmp_hdr->icmp_type, icmp_hdr->icmp_code, fstr);
+							exit_code(3);
+						}
+						else {
+							if (verbose > 2)
+								printf(": ignoring (ICMP error does not match send data)\n");
+							continue;
+						}
+					}
+					else {
+						if (verbose > 1)
+							printf(": ignoring (ICMP data is not a UDP packet)\n");
+						continue;
+					}
+				}
+#endif
 				if(ret > 0)
 				{
 					reply[ret] = 0;
 					/* store the time of our first send */
 					if (i==0)
 						memcpy(&firstsendt, &sendtime, sizeof(struct timeval));
+					retryAfter = SIP_T1;
 					/* store the biggest delay if one occured */
 					if (delaytime.tv_sec != 0) {
 						tmp_delay = deltaT(&delaytime, &recvtime);
@@ -418,26 +574,6 @@ void shoot(char *buff)
 						delaytime.tv_sec = 0;
 						delaytime.tv_usec = 0;
 					}
-					/* if (usrloc) {
-						switch (usrlocstep) {
-							case 0: 
-								cseqcmp = 3*namebeg+1;
-								break;
-							case 1:
-							case 2:
-								cseqcmp = 3*namebeg+2;
-								break;
-							case 3:
-								cseqcmp = 3*namebeg+3;
-								break;
-							default:
-								printf("error: unknown usrloc step on cseq"
-									" compare\n");
-								exit(2);
-								break;
-						}
-					}
-					else */
 					/* check for old CSeq => ignore retransmission */
 					if (!usrloc && !invite && !message)
 						cseqcmp = namebeg;
@@ -468,7 +604,7 @@ void shoot(char *buff)
 								(foo=strstr(reply, "\nm:"))==NULL ) {
 								printf("error: cannot find Contact in this "
 									"redirect:\n%s\n", reply);
-								exit(3);
+								exit_code(3);
 							}
 							crlf=strchr(foo, '\n');
 							if ((contact=strchr(foo, '\r'))!=NULL 
@@ -480,7 +616,7 @@ void shoot(char *buff)
 							if ((contact=strstr(bar, "sip"))==NULL) {
 								printf("error: cannot find sip in the Contact "
 									"of this redirect:\n%s\n", reply);
-								exit(3);
+								exit_code(3);
 							}
 							if ((foo=strchr(contact, ';'))!=NULL)
 								*foo='\0';
@@ -497,7 +633,7 @@ void shoot(char *buff)
 										printf("error: cannot handle the port "
 											"in the uri in Contact:\n%s\n", 
 											reply);
-										exit(3);
+										exit_code(3);
 									}
 								}
 								/* correct our request */
@@ -511,14 +647,14 @@ void shoot(char *buff)
 								if (!address){
 									printf("error: cannot determine host "
 										"address from Contact of redirect:"
-										"\%s\n", reply);
-									exit(2);
+										"\n%s\n", reply);
+									exit_code(2);
 								}
 							}
 							else{
 								printf("error: missing : in Contact of this "
 									"redirect:\n%s\n", reply);
-								exit(3);
+								exit_code(3);
 							}
 							free(bar);
 							memset(&addr, 0, sizeof(addr));
@@ -528,20 +664,22 @@ void shoot(char *buff)
 						else {
 							printf("error: cannot handle this redirect:"
 								"\n%s\n", reply);
-							exit(2);
+							exit_code(2);
 						}
 					} /* if redircts... */
 					else if (regexec(&authexp, reply, 0, 0, 0)==0) {
 						if (!username) {
 							printf("error: received 401 but cannot "
 								"authentication without a username\n");
-							exit(2);
+							exit_code(2);
 						}
 						/* prevents a strange error */
-						regcomp(&authexp, "^SIP/[0-9]\\.[0-9] 401 ", 
+						regcomp(&authexp, "^SIP/[0-9]\\.[0-9] 40[17] ", 
 							REG_EXTENDED|REG_NOSUB|REG_ICASE);
 						insert_auth(buff, reply);
-						i--;
+						if (verbose > 2)
+							printf("\nreceived:\n%s\n", reply);
+						increase_cseq(buff);
 					} /* if auth...*/
 					else if (trace) {
 						if (regexec(&tmhexp, reply, 0, 0, 0)==0) {
@@ -579,6 +717,7 @@ void shoot(char *buff)
 								printf("(%.3f ms) %s\n", 
 									deltaT(&sendtime, &recvtime), reply);
 							}
+							retryAfter = SIP_T2;
 							dontsend=1;
 							continue;
 						}
@@ -605,58 +744,61 @@ void shoot(char *buff)
 								printf("\twithout Contact header\n");
 							}
 							if (regexec(&okexp, reply, 0, 0, 0)==0)
-								exit(0);
+								exit_code(0);
 							else
-								exit(1);
+								exit_code(1);
 						}
 					} /* if trace ... */
 					else if (usrloc||invite||message) {
+						if (regexec(&proexp, reply, 0, 0, 0)==0) {
+							if (verbose > 2)
+								printf("\nignoring provisinal "
+									"response\n");
+							retryAfter = SIP_T2;
+							dontsend = 1;
+						}
+						else {
 						switch (usrlocstep) {
-							case 0:
+							case REG_REP:
 								/* we have sent a register and look 
 								   at the response now */
-								if (regexec(&proexp, reply, 0, 0, 0)==0) {
-									if (verbose > 2)
-										printf("\nignoring provisinal "
-											"response\n");
-									dontsend = 1;
-									break;
-								}
-								else if (regexec(&okexp, reply, 0, 0, 0)==0) {
+								if (regexec(&okexp, reply, 0, 0, 0)==0) {
 									if (verbose > 1)
 										printf ("\tOK\n");
 									if (verbose > 2)
 										printf("\n%s\n", reply);
-									//strcpy(buff, confirm);
-									//usrlocstep=1;
 								}
 								else {
 									printf("\nreceived:\n%s\nerror: didn't "
 										"received '200 OK' on register (see "
 										"above). aborting\n", reply);
-									exit(1);
+									exit_code(1);
 								}
 								if (!invite && !message) {
 									if (namebeg==nameend) {
-										printf("\nAll usrloc tests completed "
-											"successful.\nreceived last message"
-											" %.3f ms after first request (test"
-											" duration).\n", deltaT(&firstsendt,
-											 &recvtime));
-										if (big_delay)
+										if (verbose) 
+											printf("\nAll usrloc tests"
+											" completed successful.\nreceived"
+											" last message %.3f ms after first"
+											" request (test duration).\n", 
+											deltaT(&firstsendt, &recvtime));
+										if (big_delay && verbose)
 											printf("biggest delay between "
 												"request and response was %.3f"
 												" ms\n", big_delay);
-										if (retrans_r_c)
+										if (retrans_r_c && verbose)
 											printf("%i retransmission(s) "
 												"received from server.\n", 
 												retrans_r_c);
-										if (retrans_s_c)
+										if (retrans_s_c && verbose) {
 											printf("%i time(s) the timeout of "
 												"%i ms exceeded and request was"
 												" retransmitted.\n", 
 												retrans_s_c, retryAfter);
-										exit(0);
+											if (retrans_s_c > nagios_warn)
+												exit_code(4);
+										}
+										exit_code(0);
 									}
 									/* lets see if we deceid to remove a 
 									   binding (case 6)*/
@@ -678,32 +820,25 @@ void shoot(char *buff)
 										cseqcmp++;
 										trashchar=cseqcmp;
 										create_msg(buff, REQ_REM);
-										usrlocstep=6;
+										usrlocstep=UNREG_REP;
 									}
 								}
 								else if (invite) {
 									create_msg(buff, REQ_INV);
 									cseqcmp++;
-									usrlocstep=1;
+									usrlocstep=INV_RECV;
 								}
 								else if (message) {
 									create_msg(buff, REQ_MES);
 									cseqcmp++;
-									usrlocstep=4;
+									usrlocstep=MES_RECV;
 								}
 								if (sleep_ms != 0) {
 									nanosleep(&sleep_ms_s, &sleep_rem);
 								}
 								break;
-							case 1:
+							case INV_RECV:
 								/* see if we received our invite */
-								if (strncmp(reply, SIP100_STR, 
-									SIP100_STR_LEN)==0) {
-									if (verbose > 2)
-										printf("ignoring 100.. ");
-									dontsend=1;
-									continue;
-								}
 								if (!strncmp(reply, messusern, 
 									strlen(messusern))) {
 									if (verbose > 1)
@@ -713,16 +848,16 @@ void shoot(char *buff)
 									cpy_vias(reply, confirm);
 									cpy_to(reply, confirm);
 									strcpy(buff, confirm);
-									usrlocstep=2;
+									usrlocstep=INV_OK_RECV;
 								}
 								else {
 									printf("\nreceived:\n%s\nerror: did not "
 										"received the INVITE that was sent "
 										"(see above). aborting\n", reply);
-									exit(1);
+									exit_code(1);
 								}
 								break;
-							case 2:
+							case INV_OK_RECV:
 								/* did we received our ok ? */
 								if (strncmp(reply, INV_STR, INV_STR_LEN)==0) {
 									if (verbose)
@@ -739,17 +874,17 @@ void shoot(char *buff)
 										printf("\n%s\n", reply);
 									cpy_to(reply, ack);
 									strcpy(buff, ack);
-									usrlocstep=3;
+									usrlocstep=INV_ACK_RECV;
 								}
 								else {
 									printf("\nreceived:\n%s\nerror: did not "
 										"received the '200 OK' that was sent "
 										"as the reply on the INVITE (see "
 										"above). aborting\n", reply);
-									exit(1);
+									exit_code(1);
 								}
 								break;
-							case 3:
+							case INV_ACK_RECV:
 								/* did we received our ack */
 								if (nameend > 0)
 									sprintf(messusern, "%s sip:%s%i", ACK_STR, 
@@ -770,11 +905,12 @@ void shoot(char *buff)
 										printf("usrloc for %s completed "
 											"successful\n", username);
 									if (namebeg==nameend) {
-										printf("\nAll usrloc tests completed "
-											"successful.\nreceived last message"
-											" %.3f ms after first request (test"
-											" duration).\n", deltaT(&firstsendt,
-											 &recvtime));
+										if (verbose)
+											printf("\nAll usrloc tests completed "
+												"successful.\nreceived last message"
+												" %.3f ms after first request (test"
+												" duration).\n", deltaT(&firstsendt,
+												 &recvtime));
 										if (big_delay)
 											printf("biggest delay between "
 												"request and response was %.3f"
@@ -783,12 +919,15 @@ void shoot(char *buff)
 											printf("%i retransmission(s) "
 												"received from server.\n", 
 												retrans_r_c);
-										if (retrans_s_c)
+										if (retrans_s_c) {
 											printf("%i time(s) the timeout of "
 												"%i ms exceeded and request was"
 												" retransmitted.\n", 
 												retrans_s_c, retryAfter);
-										exit(0);
+											if (retrans_s_c > nagios_warn)
+												exit_code(4);
+										}
+										exit_code(0);
 									}
 									if (usrloc) {
 										/* lets see if we deceid to remove a 
@@ -800,7 +939,7 @@ void shoot(char *buff)
 											namebeg++;
 											create_msg(buff, REQ_REG);
 											cseqcmp=cseqcmp+2;
-											usrlocstep=0;
+											usrlocstep=REG_REP;
 										}
 										else {
 											/* to prevent only removing of low
@@ -812,14 +951,14 @@ void shoot(char *buff)
 											cseqcmp++;
 											trashchar=cseqcmp;
 											create_msg(buff, REQ_REM);
-											usrlocstep=6;
+											usrlocstep=UNREG_REP;
 										}
 									}
 									else {
 										namebeg++;
 										create_msg(buff, REQ_INV);
 										cseqcmp=cseqcmp+3;
-										usrlocstep=1;
+										usrlocstep=INV_RECV;
 									}
 								}
 								else {
@@ -827,12 +966,12 @@ void shoot(char *buff)
 										"received the 'ACK' that was sent "
 										"as the reply on the '200 OK' (see "
 										"above). aborting\n", reply);
-									exit(1);
+									exit_code(1);
 								}
 								if (sleep_ms != 0)
 									nanosleep(&sleep_ms_s, &sleep_rem);
 								break;
-							case 4:
+							case MES_RECV:
 								/* we sent the message and look if its 
 								   forwarded to us */
 								if (!strncmp(reply, messusern, 
@@ -848,16 +987,16 @@ void shoot(char *buff)
 									cpy_vias(reply, confirm);
 									cpy_to(reply, confirm);
 									strcpy(buff, confirm);
-									usrlocstep=5;
+									usrlocstep=MES_OK_RECV;
 								}
 								else {
 									printf("\nreceived:\n%s\nerror: did not "
 										"received the 'MESSAGE' that was sent "
 										"(see above). aborting\n", reply);
-									exit(1);
+									exit_code(1);
 								}
 								break;
-							case 5:
+							case MES_OK_RECV:
 								/* we sent our reply on the message and
 								   look if this is also forwarded to us */
 								if (strncmp(reply, MES_STR, MES_STR_LEN)==0) {
@@ -878,11 +1017,12 @@ void shoot(char *buff)
 										printf("usrloc for %s completed "
 											"successful\n", username);
 									if (namebeg==nameend) {
-										printf("\nAll usrloc tests completed "
-											"successful.\nreceived last message"
-											" %.3f ms after first request (test"
-											" duration).\n", deltaT(&firstsendt,
-											 &recvtime));
+										if (verbose)
+											printf("\nAll usrloc tests completed "
+												"successful.\nreceived last message"
+												" %.3f ms after first request (test"
+												" duration).\n", deltaT(&firstsendt,
+												 &recvtime));
 										if (big_delay)
 											printf("biggest delay between "
 												"request and response was %.3f"
@@ -891,12 +1031,15 @@ void shoot(char *buff)
 											printf("%i retransmission(s) "
 												"received from server.\n", 
 												retrans_r_c);
-										if (retrans_s_c)
+										if (retrans_s_c) {
 											printf("%i time(s) the timeout of "
 												"%i ms exceeded and request was"
 												" retransmitted.\n", 
 												retrans_s_c, retryAfter);
-										exit(0);
+											if (retrans_s_c > nagios_warn)
+												exit_code(4);
+										}
+										exit_code(0);
 									}
 									if (usrloc) {
 										/* lets see if we deceid to remove a 
@@ -908,7 +1051,7 @@ void shoot(char *buff)
 											namebeg++;
 											create_msg(buff, REQ_REG);
 											cseqcmp=cseqcmp+2;
-											usrlocstep=0;
+											usrlocstep=REG_REP;
 										}
 										else {
 											/* to prevent only removing of low
@@ -920,27 +1063,36 @@ void shoot(char *buff)
 											cseqcmp++;
 											trashchar=cseqcmp;
 											create_msg(buff, REQ_REM);
-											usrlocstep=6;
+											usrlocstep=UNREG_REP;
 										}
 									}
 									else {
 										namebeg++;
 										create_msg(buff, REQ_MES);
 										cseqcmp=cseqcmp+3;
-										usrlocstep=4;
+										usrlocstep=MES_RECV;
 									}
 								}
 								else {
-									printf("\nreceived:\n%s\nerror: did not "
-										"received the '200 OK' that was sent "
-										"as the reply on the MESSAGE (see "
-										"above). aborting\n", reply);
-									exit(1);
+									if (verbose) {
+										if (mes_body)
+											printf("\nreceived:\n%s\nerror: did"
+												" not received 200 for the "
+												"MESSAGE (see above)\n",
+												reply);
+										else
+											printf("\nreceived:\n%s\nerror: did"
+												" not received the '200 OK' "
+												"that was sent as the reply on"
+												" the MESSAGE (see above). "
+												"aborting\n", reply);
+									}
+									exit_code(1);
 								}
 								if (sleep_ms != 0)
 									nanosleep(&sleep_ms_s, &sleep_rem);
 								break;
-							case 6:
+							case UNREG_REP:
 								if (strncmp(reply, MES_STR, MES_STR_LEN)==0) {
 									if (verbose)
 										printf("ignoring MESSAGE "
@@ -961,7 +1113,7 @@ void shoot(char *buff)
 									namebeg++;
 									create_msg(buff, REQ_REG);
 									cseqcmp++;
-									usrlocstep = 0;
+									usrlocstep=REG_REP;
 									i--;
 								}
 								else {
@@ -970,16 +1122,17 @@ void shoot(char *buff)
 										"remove bindings request for %s%i (see"
 										" above). aborting\n", reply, username, 
 										namebeg);
-									exit(1);
+									exit_code(1);
 								}
 								if (sleep_ms != 0)
 									nanosleep(&sleep_ms_s, &sleep_rem);
 								break;
 							default:
 								printf("error: unknown step in usrloc\n");
-								exit(2);
+								exit_code(2);
 								break;
 						}
+						}
 					}
 					else if (randtrash) {
 						/* in randomzing trash we are expexting 4?? error codes
@@ -1007,13 +1160,13 @@ void shoot(char *buff)
 							if (randretrys == 0) {
 								printf("random end reached. server survived "
 									":) respect!\n");
-								exit(0);
+								exit_code(0);
 							}
 							else {
 								printf("maximum sendings reached but did not "
 									"get a response on this request:\n%s\n", 
 									buff);
-								exit(3);
+								exit_code(3);
 							}
 						}
 						else trash_random(buff);
@@ -1038,9 +1191,7 @@ void shoot(char *buff)
 								printf("   %s\n   provisional received; still"
 									" waiting for a final response\n", reply);
 							}
-							retryAfter = retryAfter * 2;
-							if (retryAfter > DEFAULT_TIMEOUT) 
-								retryAfter = DEFAULT_TIMEOUT;
+							retryAfter = SIP_T2;
 							dontsend = 1;
 							continue;
 						} else {
@@ -1061,9 +1212,9 @@ void shoot(char *buff)
 							}
 							else if (verbose) printf("%s\n", reply);
 							if (regexec(&okexp, reply, 0, 0, 0)==0)
-								exit(0);
+								exit_code(0);
 							else
-								exit(1);
+								exit_code(1);
 						}
 					} /* redirect, auth, and modes */
 		
@@ -1072,7 +1223,7 @@ void shoot(char *buff)
 					if (usrloc)
 						printf("failed\n");
 					perror("socket error");
-					exit(2);
+					exit_code(3);
 				}
 			} /* !flood */
 			else {
@@ -1084,7 +1235,7 @@ void shoot(char *buff)
 							deltaT(&firstsendt, &sendtime), namebeg);
 					printf("so we sended %f requests per second.\n", 
 							(namebeg/deltaT(&firstsendt, &sendtime))*1000);
-					exit(0);
+					exit_code(0);
 				}
 				namebeg++;
 				create_msg(buff, REQ_FLOOD);
@@ -1092,12 +1243,12 @@ void shoot(char *buff)
 		} /* for nretries */
 
 	} /* while redirected */
-	if (randtrash) exit(0);
+	if (randtrash) exit_code(0);
 	printf("** give up retransmissioning....\n");
 	if (retrans_r_c && (verbose > 1))
 		printf("%i retransmissions received during test\n", retrans_r_c);
 	if (retrans_s_c && (verbose > 1))
 		printf("sent %i retransmissions during test\n", retrans_s_c);
-	exit(3);
+	exit_code(3);
 }
 
